#pragma once

#ifndef LVS_EX_DOUBLEBUFFER
#define LVS_EX_DOUBLEBUFFER     0x00010000
#endif

namespace LeoHelpers
{
	inline LONG WidthRect(const RECT &rc)
	{
		return rc.right - rc.left;
	}

	inline LONG HeightRect(const RECT &rc)
	{
		return rc.bottom - rc.top;
	}


#ifdef LEOHELPERS_STRING_INCLUDED
	class RectToString
	{
	private:
		RECT m_rc;
		std::wstring m_str;
	public:
		RectToString(const RECT &rc) : m_rc(rc) {}
		static void ToString(std::wstring *pstr, const RECT &rc)
		{
			LeoHelpers::StringFormatToString(pstr, L"(L:%ld T:%ld R:%ld B:%ld W=%ld H=%ld)",
				rc.left, rc.top, rc.right, rc.bottom, WidthRect(rc), HeightRect(rc));
		}
		const wchar_t *GetString()
		{
			if (m_str.empty())
			{
				RectToString::ToString(&m_str, m_rc);
			}
			return m_str.c_str();
		}
	};
#endif

	template < typename T > class Maximizer
	{
	private:
		T m_value;
		Maximizer(const Maximizer &rhs); // disallow
		Maximizer &operator=(const Maximizer &rhs); // disallow
	public:
		Maximizer(const T &initValue) : m_value(initValue) {}
		~Maximizer() { } // non-virtual
		void Push(const T &newValue) { if (m_value < newValue) { m_value = newValue; } }
		const T &Get() const { return m_value; }
	};

	inline void AboveRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left   = pRectRelative->left;
		pRectToMove->right  = pRectToMove->left     + lWidth;
		pRectToMove->bottom = pRectRelative->top    - lOffset;
		pRectToMove->top    = pRectToMove->bottom   - lHeight;
	}

	inline void BelowRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left   = pRectRelative->left;
		pRectToMove->right  = pRectToMove->left     + lWidth;
		pRectToMove->top    = pRectRelative->bottom + lOffset;
		pRectToMove->bottom = pRectToMove->top      + lHeight;
	}

	inline void BelowRightRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->right  = pRectRelative->right;
		pRectToMove->left   = pRectToMove->right    - lWidth;
		pRectToMove->top    = pRectRelative->bottom + lOffset;
		pRectToMove->bottom = pRectToMove->top      + lHeight;
	}

	inline void RightRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left   = pRectRelative->right  + lOffset;
		pRectToMove->right  = pRectToMove->left     + lWidth;
		pRectToMove->top    = pRectRelative->top;
		pRectToMove->bottom = pRectToMove->top      + lHeight;
	}

	inline void RightBelowRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left   = pRectRelative->right  + lOffset;
		pRectToMove->right  = pRectToMove->left     + lWidth;
		pRectToMove->bottom = pRectRelative->bottom;
		pRectToMove->top    = pRectToMove->bottom   - lHeight;
	}

	inline void RightCenterRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		const LONG lRelativeHeight = pRectRelative->bottom - pRectRelative->top;

		pRectToMove->left   = pRectRelative->right  + lOffset;
		pRectToMove->right  = pRectToMove->left     + lWidth;
		pRectToMove->top    = pRectRelative->top + (lRelativeHeight - lHeight)/2;
		pRectToMove->bottom = pRectToMove->top + lHeight;
	}

	inline void LeftRect(RECT *pRectToMove, const RECT *pRectRelative, LONG lOffset)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->right  = pRectRelative->left   - lOffset;
		pRectToMove->left   = pRectToMove->right    - lWidth;
		pRectToMove->top    = pRectRelative->top;
		pRectToMove->bottom = pRectToMove->top      + lHeight;
	}

	inline void PosRectLeftTop(RECT *pRectToMove, const LONG x, const LONG y)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left = x;
		pRectToMove->right = x + lWidth;
		pRectToMove->top = y;
		pRectToMove->bottom = y + lHeight;
	}

	inline void PosRectRightTop(RECT *pRectToMove, const LONG x, const LONG y)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left = x - lWidth;
		pRectToMove->right = x;
		pRectToMove->top = y;
		pRectToMove->bottom = y + lHeight;
	}

	inline void PosRectLeftBottom(RECT *pRectToMove, const LONG x, const LONG y)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left = x;
		pRectToMove->right = x + lWidth;
		pRectToMove->top = y - lHeight;
		pRectToMove->bottom = y;
	}

	inline void PosRectRightBottom(RECT *pRectToMove, const LONG x, const LONG y)
	{
		const LONG lWidth  = pRectToMove->right  - pRectToMove->left;
		const LONG lHeight = pRectToMove->bottom - pRectToMove->top;

		pRectToMove->left = x - lWidth;
		pRectToMove->right = x;
		pRectToMove->top = y - lHeight;
		pRectToMove->bottom = y;
	}

	inline void UnionRect(RECT *pRectDest, const RECT *pRectOther)
	{
		RECT rcTemp = *pRectDest;
		::UnionRect(pRectDest, &rcTemp, pRectOther);
	}

	inline BOOL MoveWindowRect(HWND hwnd, const RECT *pRect, BOOL bRepaint)
	{
		return ::MoveWindow(hwnd, pRect->left, pRect->top, pRect->right-pRect->left, pRect->bottom-pRect->top, bRepaint);
	}

	inline BOOL ScreenToClientRect(HWND hwnd, RECT *pRect)
	{
		BOOL bResult = FALSE;

		POINT ptTL;
		POINT ptBR;

		ptTL.x = pRect->left;
		ptTL.y = pRect->top;
		ptBR.x = pRect->right;
		ptBR.y = pRect->bottom;

		if (::ScreenToClient(hwnd, &ptTL) && ScreenToClient(hwnd, &ptBR))
		{
			pRect->left   = ptTL.x;
			pRect->top    = ptTL.y;
			pRect->right  = ptBR.x;
			pRect->bottom = ptBR.y;

			bResult = TRUE;
		}

		return(bResult);
	}

	inline BOOL ClientToScreenRect(HWND hwnd, RECT *pRect)
	{
		BOOL bResult = FALSE;

		POINT ptTL;
		POINT ptBR;

		ptTL.x = pRect->left;
		ptTL.y = pRect->top;
		ptBR.x = pRect->right;
		ptBR.y = pRect->bottom;

		if (::ClientToScreen(hwnd, &ptTL) && ClientToScreen(hwnd, &ptBR))
		{
			pRect->left   = ptTL.x;
			pRect->top    = ptTL.y;
			pRect->right  = ptBR.x;
			pRect->bottom = ptBR.y;

			bResult = TRUE;
		}

		return(bResult);
	}

	// Give hWndRelative NULL to centre on the monitor work area with the mouse pointer.
	// The window is snapped to the monitor work area it appears on.
	void CenterWindow(HWND hWndToMove, HWND hWndRelative);

	// Reduces the window's size if it's too large to fit on the workspace of the monitor it is
	// most on. Returns true if any shrinking was performed. Will not shrink the window smaller
	// than allowed by WM_GETMINMAXINFO.
	bool FitToWorkspace(HWND hWnd);

	inline void CenterAndFitToWorkspace(HWND hWndToMove, HWND hWndRelative)
	{
		LeoHelpers::CenterWindow(hWndToMove, hWndRelative); // Center once to get it on the correct monitor.

		if (LeoHelpers::FitToWorkspace(hWndToMove))
		{
			LeoHelpers::CenterWindow(hWndToMove, hWndRelative); // Center again if it was resized.
		}
	}

	// Flags as for MonitorFromWindow.
	// If you don't want a rect pass in NULL.
	bool GetAreasFromWindow(RECT *pRectFull, RECT *pRectWork, HWND hWnd, DWORD dwFlags);

	// DeleteObject the result (if non-NULL)
//	HFONT GetMessageFont();

	class DialogLayoutHelper
	{
	private:
		HWND m_hwndDlg;
		LONG m_lBaseX;
		LONG m_lBaseY;
		RECT m_rect1;
		RECT m_rect2;
		RECT m_rect3;
		RECT m_rect4;

	public:
		DialogLayoutHelper(HWND hwndDlg); // This is preferable if you have a dialog, but won't work with non-dialog windows.
		DialogLayoutHelper(HWND hwndControl, HFONT hFont); // If hFont is NULL, WM_GETFONT will be sent to hwndControl.
		DialogLayoutHelper(const DialogLayoutHelper &rhs) = default;
		DialogLayoutHelper &operator=(const DialogLayoutHelper &rhs) = default;

		// non-virtual
		~DialogLayoutHelper() { }

	private:

		void CommonInit();

		// Private because a fudge-factor is applied by the EnsureComboBoxWidth wrapper, so things should always use the wrapper instead.
		static LONG GetComboBoxItemWidth(HWND hWndCombo);

	public:
		BOOL MapDialogRect(RECT *pRect) const;

		LONG GetRelatedControlGapXPixels()    const { return m_rect1.left;   }
		LONG GetRelatedControlGapYPixels()    const { return m_rect1.top;    }
		LONG GetButtonWidthXPixels()          const { return m_rect3.left;   }
		LONG GetButtonGapXPixels()            const { return m_rect1.left;   }
		LONG GetButtonGapYPixels()            const { return m_rect1.top;    }
		LONG GetMarginPixels()                const { return m_rect1.bottom; }
		LONG GetDialogMarginPixels()          const { return (m_rect1.bottom+1)/2; }

		LONG GetLabelAssocGapYPixels()        const { return m_rect2.top;    }

		LONG GetGroupBoxFirstControlYPixels() const { return m_rect3.top; }

		static bool MeasureTextSingleLine(HWND hWndControl, const wchar_t *szText, RECT *prc, bool bNoPrefix);
		static LONG MeasureTextSingleLineWidth(HWND hWndControl, const wchar_t *szText, bool bNoPrefix)
		{
			RECT rc = {0};
			if (MeasureTextSingleLine(hWndControl, szText, &rc, bNoPrefix))
			{
				return rc.right - rc.left;
			}
			return 0;
		}

		// Calling this (or otherwise moving/sizing the combo) will zero the drop-down list's height on XP.
		// See EnsureComboBoxListHeight, below, for more info and how to fix it.
		bool EnsureComboBoxWidth(HWND hWndCombo, LONG lMaxLabelWidth, bool bResizeWindow, RECT *prcComboOut) const;

		// MUST be called on Windows XP, else the list will have zero height. (Vista and above are more intelligent.)
		// MUST be called *AFTER* any other moving or sizing of the combo, else the stupid thing gets undone on XP.
		// (If you get the combo control's rect it'll tell you the rect of the actual control. If you set the rect
		// then the height defines that of the drop-down list (NOT the control!). So just getting the rect, adjusting
		// the position and/or width (not even touching the height!) and then setting the rect will make the drop-down
		// zero height on XP. Sigh of bloody sighs.)
		bool EnsureComboBoxListHeight(HWND hWndCombo, int iMinLineCountOrZero) const;

		// Resize static only pays attention to the input of pInOutRect if bSingleLine is false.
		bool ResizeStatic(  HWND hwndStatic,   RECT *pInOutRect, bool bResize, bool bRedraw, bool bSingleLine) const;
		bool ResizeProgress(HWND hwndProgress, RECT *pOutRect, bool bResize, bool bRedraw) const;
		bool ResizeCheckbox(HWND hwndCheckbox, RECT *pOutRect, bool bResize, bool bRedraw) const;
		bool ResizeEdit(    HWND hwndEdit,     RECT *pOutRect, bool bResize, bool bRedraw, bool bPreserveCurrentWidth, bool bMinimumHeight, bool bLargerForSpin) const;
		bool ResizeButton(  HWND hwndButton,   RECT *pOutRect, bool bResize, bool bRedraw, bool bIgnoreMinimumWidth, LONG lFixedHeight) const;

		static LONG GetEditMinimumHeight(HWND hwndEdit);

		void SetGroupBoxText(HWND hwndGroupBox, const wchar_t *szText, bool bNoPrefix) const;

		HWND GetHwndDlg() const { return m_hwndDlg; } // Returns NULL if the constructor taking a control was used instead of the dialog one.
	};

	class WindowMaker
	{
	protected:
		HMODULE m_hModule;
		HWND m_hWnd;
		DialogLayoutHelper *m_pLayout;
	public:
		WindowMaker(HMODULE hModule, HWND hWnd, DialogLayoutHelper *pLayout)
			: m_hModule(hModule)
			, m_hWnd(hWnd)
			, m_pLayout(pLayout)
		{
		}

		virtual ~WindowMaker()
		{
		}

	private:
		WindowMaker(const WindowMaker &rhs); // disallow
		WindowMaker &operator=(const WindowMaker &rhs); // disallow

	public:
		bool CreateCheck(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, const wchar_t *szLabel, bool bResize, bool bRadio, bool bTriState, bool bAuto, bool bTabStop) const;
		bool CreateEdit(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, bool bResize, LONG lWidth, bool bAutoHScroll, bool bRight, bool bNumber, bool bReadOnly, bool bPassword, bool bLargerForSpin) const;
		bool CreateSpin(HWND *pOutHwnd, ULONG uiCtrlId, bool bSetBuddyInt, bool bRight, bool bArrowKeys, bool bNoThousands) const;
		bool CreateGroup(HWND *pOutHwnd, ULONG uiCtrlId, const wchar_t *szLabel) const;
		bool CreateButton(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, const wchar_t *szLabel, bool bResize, bool bDefPushButton) const;
		bool CreateListBox(HWND *pOutHwnd, ULONG uiCtrlId, bool bSort, bool bExtendedSel) const;
		bool CreateListViewReport(HWND *pOutHwnd, ULONG uiCtrlId, bool bSortAscending, bool bSortDescending, bool bNoColumnHeader, bool bNoSortHeader, bool bShowSelAlways, bool bSingleSel, bool bFullRowSelect, bool bEditLabels, bool bCheckboxes, bool bShareImageLists) const;
		bool CreateLabel(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, const wchar_t *szLabel, bool bResize, bool bSingleLine, bool bVCenter) const;
		bool CreateProgress(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, bool bResize, bool bMarquee, bool bSmooth, bool bSmoothReverse) const;
		bool CreateBitmap(HWND *pOutHwnd, ULONG uiCtrlId, bool bNotify) const;
		bool CreateCombo(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, bool bResize, bool bDropDownList, bool bDropDown) const;
		bool CreateTab(HWND *pOutHwnd, ULONG uiCtrlId) const;
		bool CreateResizeGrip(HWND *pOutHwnd, ULONG uiCtrlId) const;
		bool CreateTabChildDlg(HWND *pOutHwnd, ULONG uiDlgRes, HWND hWndTab, WPARAM &tabNum, const wchar_t *szLabel, DLGPROC dlgProc, LPARAM lTabParam) const;
	};

#ifdef DOPUS_PLUGIN_HELPER

	class OpusWindowMaker : public WindowMaker
	{
	private:
		OpusStringLoader *m_pStringLoader;
	public:
		OpusWindowMaker(HMODULE hModule, HWND hWnd, DialogLayoutHelper *pLayout, OpusStringLoader *pStringLoader)
			: WindowMaker(hModule, hWnd, pLayout)
			, m_pStringLoader(pStringLoader)
		{
		}

		virtual ~OpusWindowMaker()
		{
		}

	private:
		OpusWindowMaker(const OpusWindowMaker &rhs); // disallow
		OpusWindowMaker &operator=(const OpusWindowMaker &rhs); // disallow

	public:
		// "OS" for "Opus String" is added to method names, rather than overloading them, to avoid overloads which only differ on UINT / const wchar_t *
		// arguments, which could lead to problems with ambiguous arguments (e.g. NULL).

		inline OpusStringLoader *GetOpusStringLoader() const { return m_pStringLoader; }

		inline bool CreateCheckOS(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, UINT uiLabel, bool bResize, bool bRadio, bool bTriState, bool bAuto, bool bTabStop) const
		{
			const wchar_t *szLabel = ((uiLabel == 0) ? L"" : m_pStringLoader->Get(uiLabel));
			return this->CreateCheck(pOutHwnd, pOutRect, uiCtrlId, szLabel, bResize, bRadio, bTriState, bAuto, bTabStop);
		}

		inline bool CreateGroupOS(HWND *pOutHwnd, ULONG uiCtrlId, UINT uiLabel) const
		{
			const wchar_t *szLabel = ((uiLabel == 0) ? L"" : m_pStringLoader->Get(uiLabel));
			return this->CreateGroup(pOutHwnd, uiCtrlId, szLabel);
		}

		inline bool CreateButtonOS(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, UINT uiLabel, bool bResize, bool bDefPushButton) const
		{
			const wchar_t *szLabel = ((uiLabel == 0) ? L"" : m_pStringLoader->Get(uiLabel));
			return this->CreateButton(pOutHwnd, pOutRect, uiCtrlId, szLabel, bResize, bDefPushButton);
		}

		inline bool CreateLabelOS(HWND *pOutHwnd, RECT *pOutRect, ULONG uiCtrlId, UINT uiLabel, bool bResize, bool bSingleLine, bool bVCenter) const
		{
			const wchar_t *szLabel = ((uiLabel == 0) ? L"" : m_pStringLoader->Get(uiLabel));
			return this->CreateLabel(pOutHwnd, pOutRect, uiCtrlId, szLabel, bResize, bSingleLine, bVCenter);
		}

		inline bool CreateTabChildDlgOS(HWND *pOutHwnd, ULONG uiDlgRes, HWND hWndTab, WPARAM &tabNum, UINT uiLabel, DLGPROC dlgProc, LPARAM lTabParam) const
		{
			const wchar_t *szLabel = ((uiLabel == 0) ? L"" : m_pStringLoader->Get(uiLabel));
			return this->CreateTabChildDlg(pOutHwnd, uiDlgRes, hWndTab, tabNum, szLabel, dlgProc, lTabParam);
		}

		// dwFlags should normally be set to GAFDT_WANT_SEGOE
		bool CreateTabChildOpusDlg(HWND *pOutHwnd, ULONG uiDlgRes, HWND hWndTab, WPARAM &tabNum, const wchar_t *szLabel, DLGPROC dlgProc, LPARAM lTabParam, DWORD dwFlags) const;

		// dwFlags should normally be set to GAFDT_WANT_SEGOE
		inline bool CreateTabChildOpusDlgOS(HWND *pOutHwnd, ULONG uiDlgRes, HWND hWndTab, WPARAM &tabNum, UINT uiLabel, DLGPROC dlgProc, LPARAM lTabParam, DWORD dwFlags) const
		{
			const wchar_t *szLabel = ((uiLabel == 0) ? L"" : m_pStringLoader->Get(uiLabel));
			return this->CreateTabChildOpusDlg(pOutHwnd, uiDlgRes, hWndTab, tabNum, szLabel, dlgProc, lTabParam, dwFlags);
		}
	};

#endif

};
