#include "stdafx.h"
#include "LeoHelpers_String.h"
#include "LeoHelpers_Version.h"

bool LeoHelpers::CopyVersionNumber(DWORD *pdwHigh, DWORD *pdwLow, VERNUM_TYPE verType, const LPVOID pVerData)
{
	bool bResult = false;

	LPVOID pBuffer = NULL;
	UINT bufLen = 0;

	if (::VerQueryValue(pVerData, L"\\", &pBuffer, &bufLen)
	&&	bufLen >= sizeof(VS_FIXEDFILEINFO))
	{
		const VS_FIXEDFILEINFO *pFileInfo = reinterpret_cast< const VS_FIXEDFILEINFO * >( pBuffer );
		switch(verType)
		{
		default:
			assert(false);
			break;
		case VNT_FILEVERSION:
			*pdwHigh = pFileInfo->dwFileVersionMS;
			*pdwLow  = pFileInfo->dwFileVersionLS;
			bResult = true;
			break;
		case VNT_PRODUCTVERSION:
			*pdwHigh = pFileInfo->dwFileVersionMS;
			*pdwLow  = pFileInfo->dwFileVersionLS;
			bResult = true;
			break;
		}
	}

	return(bResult);
}

bool LeoHelpers::CopyVersionString(wchar_t *szDestBuffer, UINT uiDestBufferSize, wchar_t *szItemName, BYTE *pLang, const LPVOID pVerData)
{
	bool bResult = false;

	if (NULL == szDestBuffer)
	{
		bResult = true;
	}
	else if (uiDestBufferSize > 2)
	{
		wchar_t szVerPath[128];
		swprintf_s(szVerPath, 128, L"\\StringFileInfo\\%02x%02x%02x%02x\\%s", pLang[1], pLang[0], pLang[3], pLang[2], szItemName);

		LPVOID pBuffer;
		UINT bufLen;

		if (::VerQueryValue(pVerData, szVerPath, &pBuffer, &bufLen))
		{
			wcsncpy_s(szDestBuffer, uiDestBufferSize, reinterpret_cast< wchar_t * >(pBuffer), _TRUNCATE);

			bResult = true;
		}
	}

	return(bResult);
}


// Give NULL binary path to get the version of the hModule (hModule is unused otherwise).
// Give NULL hModule as well to get the version of the current exe.
// pstrOut or pVerNumOut can be NULL; just give the one(s) you want.
bool LeoHelpers::GetVersionNumber(DWORD *pdwHigh, DWORD *pdwLow, const wchar_t *szBinaryPath, HMODULE hModule)
{
	wchar_t szModNameBuf[_MAX_PATH + 1];

	if (szBinaryPath == NULL)
	{
		if (0 == ::GetModuleFileName(hModule, szModNameBuf, _MAX_PATH))
		{
			return false;
		}

		szBinaryPath = szModNameBuf;
	}

	bool bResult = false;
	DWORD dwVerSize;
	DWORD dwUnused;

	if (0 != (dwVerSize = ::GetFileVersionInfoSize(szBinaryPath, &dwUnused)))
	{
		BYTE *pVerData = new BYTE[dwVerSize];

		if (::GetFileVersionInfo(szBinaryPath, NULL, dwVerSize, pVerData))
		{
			if (LeoHelpers::CopyVersionNumber(pdwHigh, pdwLow, VNT_FILEVERSION, pVerData))
			{
				bResult = true;
			}
		}

		delete [] pVerData;
	}

	return bResult;
}
